// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <string.h>

#include "view_ss.h"
#include "http_server.h"
#include "utils.h"
#include "picio.h"
#include "log.h"

view_ss::view_ss(configuration_t *const cfg, const std::string & id, const std::string & descr, const int width, const int height, const bool auto_shrink, resize *const r, const std::vector<view_src_t> & sources, const double switch_interval, std::vector<target *> *const targets, std::vector<filter *> *const filters) : view(cfg, id, descr, width, height, r, sources, filters), auto_shrink(auto_shrink), switch_interval(switch_interval), targets(targets)
{
	cur_source_index = 0;
        latest_switch = 0;
	cur_source = this;
}

view_ss::~view_ss()
{
	if (targets) {
		for(auto t : *targets)
			delete t;

		delete targets;
	}

	free(prev_frame);
}

void view_ss::operator()()
{
	log(LL_INFO, "Starting view_ss thread");

	if (targets) {
		log(LL_INFO, "Starting %zu targets", targets -> size());

		struct timeval ets;
		gettimeofday(&ets, nullptr);
		unsigned long int event_nr = get_db() -> register_event(id, EVT_GENERIC, "FIXME", &ets);

		for(auto & t : *targets)
			t->start(nullptr, event_nr);
	}

	log(LL_INFO, "Stopping view_ss thread");
}

std::string view_ss::get_html(const std::map<std::string, std::string> & pars) const
{
	std::string dim_img;

	if (width != -1)
		dim_img += myformat(" width=%d", width);
	if (height != -1)
		dim_img += myformat(" height=%d", height);

	return myformat("<html><head><style></style></head><body><img src=\"stream.mjpeg?int=%s&view-proxy\"%s></body></html>", url_escape(id).c_str(), url_escape(dim_img).c_str());
}

void view_ss::register_user()
{
	user_count++;

	for(auto s : sources) {
		instance_t *inst = nullptr;
		source *source = nullptr;
		find_by_id(cfg, s.id, &inst, (interface **)&source);

		if (source)
			source -> register_user();
	}
}

void view_ss::unregister_user()
{
	user_count--;

	for(auto s : sources) {
		instance_t *inst = nullptr;
		source *source = nullptr;
		find_by_id(cfg, s.id, &inst, (interface **)&source);

		if (source)
			source -> unregister_user();
	}
}

source *view_ss::get_current_source()
{
	return cur_source;
}

bool view_ss::get_frame(const encoding_t pe, const int jpeg_quality, uint64_t *ts, int *o_width, int *o_height, uint8_t **frame, size_t *frame_len, const bool use_previous_when_fail)
{
	instance_t *inst = nullptr;
	find_by_id(cfg, sources.at(cur_source_index).id, &inst, (interface **)&cur_source);

	*ts = 0;

	uint64_t si = switch_interval * 1000000.0;
	if (switch_interval < 1.0)
		si = 1000000;

	if (latest_switch == 0)
		latest_switch = get_us();

	uint64_t now = get_us();
	if (now - latest_switch >= si) {
		for(size_t i=0; i<sources.size(); i++) {
			if (++cur_source_index >= sources.size())
				cur_source_index = 0;

			instance_t *next_inst = nullptr;
			interface *next_source = nullptr;
			find_by_id(cfg, sources.at(cur_source_index).id, &next_inst, &next_source);

			if (!next_source)
				log(LL_ERR, "source selection %s in view %s is not known", sources.at(cur_source_index).id.c_str(), id.c_str());
			else if (next_source->is_paused() == false && next_source->is_running() == true)
				break;
		}

		latest_switch = now;
	}

	*o_width = *o_height = -1;

	if (!cur_source)
		return false;

	bool rc = false;

	if (auto_shrink || !filters->empty()) {
		int curw = -1, curh = -1;
		uint8_t *curf = nullptr;
		size_t curlen = 0;

		cur_source -> register_user();
		rc = cur_source -> get_frame(E_RGB, 0, ts, &curw, &curh, &curf, &curlen, use_previous_when_fail);
		cur_source -> unregister_user();

		if (rc) {
			if (width == -1)
				width = curw;
			if (height == -1)
				height = curh;

			if (!prev_frame)
				prev_frame = (uint8_t *)allocate_aligned(width * height * 3);

			uint8_t *resized = nullptr;

			r -> do_resize(curw, curh, curf, width, height, &resized);
			free(curf);

			if (filters) {
				apply_filters(inst, this, filters, prev_frame, resized, *ts, width, height);
				memcpy(prev_frame, resized, width * height * 3);
			}

			if (pe == E_RGB) {
				*frame = resized;
				*frame_len = width * height * 3;
			}
			else if (pe == E_JPEG) {
				rc = write_JPEG_memory(get_meta(), width, height, jpeg_quality, resized, (char **)frame, frame_len);
			}

			*o_width = width;
			*o_height = height;
		}
	}
	else {
		cur_source -> register_user();
		rc = cur_source -> get_frame(pe, jpeg_quality, ts, o_width, o_height, frame, frame_len, use_previous_when_fail);
		cur_source -> unregister_user();
	}

	return rc;
}

void view_ss::stop()
{
	log(LL_INFO, "view_ss::stop");

	if (targets) {
		for(target *t : *targets) {
			t -> stop();
			delete t;
		}

		targets -> clear();
	}

	interface::stop();
}
